@@ -1,8 +1,7 @@
 import random
-
 import requests
 
-from .data import string_to_raw, time_to_raw, float_to_raw
+from .data import string_to_raw, float_to_raw, int_to_raw, bool_to_raw
 from .stream import Stream
 
 
@@ -83,6 +82,8 @@ class Client:
         return obj
 
     def create_object(self, abstracts, name='', space=None, dependent=False):
+        if not isinstance(abstracts, list):
+            abstracts = [abstracts]
         if not space:
             if not self.__space:
                 raise RuntimeError('no universe or space selected')
@@ -95,7 +96,9 @@ class Client:
             'dependent': dependent,
         })['object']
 
-    def create_relation(self, origin_id, target_id, abstracts, name='', space=None, time=None):
+    def create_relation(self, origin_id, target_id, abstracts, name='', space=None):
+        if not isinstance(abstracts, list):
+            abstracts = [abstracts]
         if not space:
             if not self.__space:
                 raise RuntimeError('no universe or space selected')
@@ -109,9 +112,6 @@ class Client:
             'name': name
         }
 
-        if time is not None:
-            body['time'] = time.isoformat() + 'Z'
-
         return self.__post(f'{self.__system_url}relations', body)['relation']
 
     def get_object(self, id: str):
@@ -120,6 +120,10 @@ class Client:
     def delete_object(self, id: str):
         return self.__delete(f'{self.__system_url}objects/{self.resolve(id)}')
 
+    def delete_objects(self, ids: list):
+        for id in ids:
+            self.delete_object(id)
+
     def set_name(self, id: str, name: str):
         self.__post(f'{self.__system_url}objects/{self.resolve(id)}/name', {
             'name': name,
@@ -148,7 +152,7 @@ class Client:
     def get_abstracts(self, id: str, include_self=False, include_direct=True, include_transitive=False):
         return self.__get_hierarchy(self.resolve(id), include_self, include_direct, include_transitive, 'abstracts', self.__abstract_cache)
 
-    def get_specials(self, id: str, include_self=False, include_direct=True, include_transitive=False):
+    def get_instances(self, id: str, include_self=False, include_direct=True, include_transitive=False):
         return self.__get_hierarchy(self.resolve(id), include_self, include_direct, include_transitive, 'specials', self.__special_cache)
 
     def get_targets(self, id: str) -> dict:
@@ -157,23 +161,38 @@ class Client:
     def get_origins(self, id: str) -> dict:
         return self.__get_relations(self.resolve(id), 'origins', self.__origins_cache)
 
+    def clear_values(self, id: str, field: str):
+        field_id = self.resolve(field)
+        query, abbr, _ = self.__get_values_query(field_id, 'val', 'id')
+        vals = self.query(query, {'obj': {'id': id}, abbr: {'id': field_id}})
+        for v in vals:
+            self.delete_object(v['rval'])
+
     def set_value(self, id: str, field: str, value):
+        self.clear_values(id, field)
         self.add_value(id, field, value)
 
-    def add_value(self, id: str, field: str, value, time=None):
+    def add_value(self, id: str, field: str, value):
         field_id = self.resolve(field)
         targets = self.get_targets(field_id)
         target_id = targets['ffffffffffffffffffffffffffffffff']
-        data = self.__get_type(target_id)
-        if data == 'string':
+        data_type = self.__get_type(target_id)
+        data = None
+        if data_type == 'string':
             data = string_to_raw(value)
-        elif data == 'float':
+        elif data_type == 'float':
             data = float_to_raw(value)
-        elif data == 'time':
-            data = time_to_raw(value)
-        target = self.create_object([target_id], dependent=True)
-        self.set_data(target['id'], data)
-        r = self.create_relation(id, target['id'], [field_id], time=time)
+        elif data_type == 'int':
+            data = int_to_raw(value)
+        elif data_type == 'bool':
+            data = bool_to_raw(value)
+        if data:
+            target = self.create_object([target_id], dependent=True)
+            target_id = target['id']
+            self.set_data(target_id, data)
+            r = self.create_relation(id, target_id, [field_id])
+        else:
+            r = self.create_relation(id, value, [field_id])
         return r
 
     def get_value(self, id: str, field: str) -> any:
@@ -182,7 +201,7 @@ class Client:
 
     def get_values(self, id: str, field: str):
         field_id = self.resolve(field)
-        query, abbr = self.__get_values_query(field_id, 'val')
+        query, abbr, data_type = self.__get_values_query(field_id, 'val')
         vals = self.query(query, {'obj': {'id': id}, abbr: {'id': field_id}})
         return [v['val'] for v in vals]
 
@@ -192,7 +211,7 @@ class Client:
         v = {}
         for name in names:
             field_id = self.resolve(defs[name])
-            frag, abbr = self.__get_values_query(field_id, name)
+            frag, abbr, _ = self.__get_values_query(field_id, name)
             query += frag
             v[abbr] = {'id': field_id}
         v['obj'] = {'id': id}
@@ -207,6 +226,35 @@ class Client:
         for key, val in items:
             self.set_value(id, defs[key], val)
 
+    def lookup_value(self, field: str, value):
+        field_id = self.resolve(field)
+        targets = self.get_targets(field_id)
+        target_id = targets['ffffffffffffffffffffffffffffffff']
+        data_type = self.__get_type(target_id)
+        raw = None
+        if isinstance(value, str) and len(value) == 32:
+            query = \
+                f'value ~ $target;' \
+                f'value === <{value}>;' \
+                f'origin.[_:$rel] == value;' \
+                f'id(origin) => val;'
+        else:
+            if data_type == 'string':
+                raw = f'"{value}"'
+            elif data_type == 'int':
+                raw = f'{value:.0d}'
+            elif data_type == 'float':
+                raw = f'{value:f}'
+            elif data_type == 'bool':
+                raw = 'true' if value else 'false'
+            query = \
+                f'value ~ $target;' \
+                f'value == {raw};' \
+                f'origin.[_:$rel] == value;' \
+                f'id(origin) => val;'
+        vals = self.query(query, {'target': {'id': target_id}, 'rel': {'id': field_id}})
+        return [v['val'] for v in vals]
+
     def stream(self) -> Stream:
         return Stream(self, self.__stream_url)
 
@@ -216,13 +264,16 @@ class Client:
             'valuation': valuation,
         })['valuations']
 
-    def __get_values_query(self, field_id, name):
+    def __get_values_query(self, field_id, name, data_type=None):
         targets = self.get_targets(field_id)
         target_id = targets['ffffffffffffffffffffffffffffffff']
-        data = self.__get_type(target_id)
+        if not data_type:
+            data_type = self.__get_type(target_id)
+        if not data_type:
+            data_type = 'id'
         abbr = 'A' + field_id[:6]
-        return f'$obj.[_:${abbr}] == {abbr};' \
-               f'{data}({abbr}) => {name};', abbr
+        return f'$obj.[r{abbr}:${abbr}] == {abbr};' \
+               f'{data_type}({abbr}) => {name};id(r{abbr}) => r{name};', abbr, data_type
 
     def __get_type(self, id: str):
         abs = self.get_abstracts(id, True, True, True)
@@ -233,9 +284,7 @@ class Client:
         if self.resolve('data.@integer') in abs:
             return 'int'
         if self.resolve('data.@boolean') in abs:
-            return 'boolean'
-        if self.resolve('data.@time') in abs:
-            return 'time'
+            return 'bool'
         return None
 
     def __get_hierarchy(self, id: str, include_self: bool, include_direct: bool, include_transitive: bool, name: str, cache: dict):
@@ -254,9 +303,10 @@ class Client:
         ck = f'{id}'
         rel = cache.get(ck)
         if not rel:
-            rel = self.__get(
-                f'{self.__system_url}objects/{id}/{name}')['idPairs']
-            cache[ck] = rel
+            resp = self.__get(f'{self.__system_url}objects/{id}/{name}')
+            if resp:
+                rel = resp['idPairs']
+                cache[ck] = rel
         return rel
 
     def __get(self, url, is_json=True, **kwargs):
@@ -323,7 +373,7 @@ class Client:
             return self.__refresh()
 
         perm_reqs = req.split('/')
-        if perm_reqs[0] == '00000000000000000000000000000000':
+        if perm_reqs[0] == '00000000000000000000000000000000' or perm_reqs[0] == '':
             return False
 
         tk = self.__space_tokens.get(perm_reqs[0])
